from __future__ import annotations

import os
import re
import hashlib
import inspect
from typing import Any, Literal, Mapping, Callable, Optional, Awaitable, cast
from dataclasses import field, replace, dataclass
from typing_extensions import Self, override

import httpx

from ..auth import WorkloadIdentity
from .._types import NOT_GIVEN, Timeout, NotGiven
from .._utils import is_given
from .._client import OpenAI, AsyncOpenAI
from .._models import FinalRequestOptions
from .._provider import _Provider, _configure_provider
from .._exceptions import OpenAIError
from .._base_client import DEFAULT_MAX_RETRIES
from ..providers.bedrock import AwsCredentialsProvider, bedrock, _BedrockProviderRuntime

BedrockTokenProvider = Callable[[], str]
AsyncBedrockTokenProvider = Callable[[], "str | Awaitable[str]"]
_LegacyAuthMode = Literal["bearer", "token_provider", "aws"]
_LegacyAuthConfiguration = tuple[_LegacyAuthMode, Optional[object]]
_LEGACY_SIGNATURE_KEY = os.urandom(32)


@dataclass(frozen=True)
class _LegacyRuntimeSignature:
    mode: _LegacyAuthMode
    base_url: str
    region: str | None
    credential_identity: object = field(repr=False)


@dataclass(frozen=True)
class _LegacyBedrockState:
    explicit_api_key: str | None = field(repr=False)
    token_provider: BedrockTokenProvider | AsyncBedrockTokenProvider | None = field(repr=False, compare=False)
    aws_region: str | None
    region_was_explicit: bool
    aws_profile: str | None
    aws_access_key_id: str | None = field(repr=False)
    aws_secret_access_key: str | None = field(repr=False)
    aws_session_token: str | None = field(repr=False)
    aws_credentials_provider: AwsCredentialsProvider | None = field(repr=False, compare=False)
    uses_environment_bearer: bool
    environment_bearer_token: str | None = field(repr=False)
    uses_region_derived_base_url: bool


def _state_api_key(state: _LegacyBedrockState) -> str:
    return state.explicit_api_key or (state.environment_bearer_token if state.uses_environment_bearer else "") or ""


def _constructor_accepts_keyword(constructor: Callable[..., object], name: str) -> bool:
    try:
        parameters = inspect.signature(constructor).parameters
    except (TypeError, ValueError):
        return False

    return name in parameters or any(
        parameter.kind is inspect.Parameter.VAR_KEYWORD for parameter in parameters.values()
    )


def _configured_region(region: str | None) -> str | None:
    configured = region or os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION")
    return configured.strip() if configured is not None and configured.strip() else None


def _uses_region_derived_base_url(base_url: str | httpx.URL | None) -> bool:
    if isinstance(base_url, str) and not base_url.strip():
        base_url = None
    if base_url is not None:
        return False

    environment_base_url = os.environ.get("AWS_BEDROCK_BASE_URL")
    return environment_base_url is None or not environment_base_url.strip()


def _has_explicit_aws_auth(
    *,
    aws_profile: str | None,
    aws_access_key_id: str | None,
    aws_secret_access_key: str | None,
    aws_session_token: str | None,
    aws_credentials_provider: AwsCredentialsProvider | None,
) -> bool:
    return any(
        value is not None
        for value in (
            aws_profile,
            aws_access_key_id,
            aws_secret_access_key,
            aws_session_token,
            aws_credentials_provider,
        )
    )


def _environment_bearer_token() -> str:
    token = os.environ.get("AWS_BEARER_TOKEN_BEDROCK")
    if not token:
        raise OpenAIError(
            "Could not find credentials for Bedrock. Set `AWS_BEARER_TOKEN_BEDROCK` or configure the default "
            "AWS credential chain."
        )
    return token


def _legacy_provider(
    *,
    api_key: str | None,
    token_provider: BedrockTokenProvider | AsyncBedrockTokenProvider | None,
    aws_region: str | None,
    aws_profile: str | None,
    aws_access_key_id: str | None,
    aws_secret_access_key: str | None,
    aws_session_token: str | None,
    aws_credentials_provider: AwsCredentialsProvider | None,
    base_url: str | httpx.URL | None,
    region_was_explicit: bool | None = None,
) -> tuple[_Provider, _LegacyBedrockState, str]:
    if callable(cast(object, api_key)):
        raise OpenAIError("Pass refreshable Bedrock credentials via `bedrock_token_provider`, not `api_key`.")
    if api_key == "":
        raise OpenAIError("The `api_key` argument must not be empty.")
    if api_key is not None and token_provider is not None:
        raise OpenAIError(
            "Bedrock authentication is ambiguous. Configure exactly one explicit mode: bearer credential, "
            "static AWS credentials, profile, or credential provider."
        )

    explicit_aws_auth = _has_explicit_aws_auth(
        aws_profile=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        aws_credentials_provider=aws_credentials_provider,
    )
    if (api_key is not None or token_provider is not None) and explicit_aws_auth:
        raise OpenAIError(
            "Bedrock authentication is ambiguous. Configure exactly one explicit mode: bearer credential, "
            "static AWS credentials, profile, or credential provider."
        )

    environment_token = os.environ.get("AWS_BEARER_TOKEN_BEDROCK")
    uses_environment_bearer = (
        api_key is None and token_provider is None and not explicit_aws_auth and bool(environment_token)
    )
    resolved_region = _configured_region(aws_region)
    uses_region_derived_base_url = _uses_region_derived_base_url(base_url)

    provider_base_url: str | httpx.URL | None | NotGiven
    if isinstance(base_url, str) and not base_url.strip():
        provider_base_url = None
    elif base_url is None:
        provider_base_url = NOT_GIVEN
    else:
        provider_base_url = base_url

    provider = bedrock(
        region=aws_region,
        base_url=provider_base_url,
        api_key=api_key if api_key is not None else environment_token if uses_environment_bearer else NOT_GIVEN,
        token_provider=token_provider,
        access_key_id=aws_access_key_id,
        secret_access_key=aws_secret_access_key,
        session_token=aws_session_token,
        profile=aws_profile,
        credential_provider=aws_credentials_provider,
    )
    state = _LegacyBedrockState(
        explicit_api_key=api_key,
        token_provider=token_provider,
        aws_region=resolved_region,
        region_was_explicit=(
            bool(aws_region and aws_region.strip()) if region_was_explicit is None else region_was_explicit
        ),
        aws_profile=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        aws_credentials_provider=aws_credentials_provider,
        uses_environment_bearer=uses_environment_bearer,
        environment_bearer_token=environment_token if uses_environment_bearer else None,
        uses_region_derived_base_url=uses_region_derived_base_url,
    )
    return provider, state, api_key or (environment_token if uses_environment_bearer else "") or ""


def _copy_configuration(
    client: BedrockOpenAI | AsyncBedrockOpenAI,
    *,
    api_key: str | None,
    token_provider: BedrockTokenProvider | AsyncBedrockTokenProvider | None,
    aws_region: str | None,
    aws_profile: str | None,
    aws_access_key_id: str | None,
    aws_secret_access_key: str | None,
    aws_session_token: str | None,
    aws_credentials_provider: AwsCredentialsProvider | None,
    base_url: str | httpx.URL | None,
) -> tuple[dict[str, object], _Provider | None, _LegacyBedrockState | None]:
    _synchronize_legacy_routing_state(client)
    state = client._bedrock_state
    current_api_key = client.api_key or ""
    api_key_was_mutated = state.token_provider is None and current_api_key != _state_api_key(state)
    aws_override = _has_explicit_aws_auth(
        aws_profile=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        aws_credentials_provider=aws_credentials_provider,
    )
    explicit_bearer_override = api_key is not None or token_provider is not None
    if explicit_bearer_override and aws_override:
        raise OpenAIError(
            "Bedrock authentication is ambiguous. Configure exactly one explicit mode: bearer credential, "
            "static AWS credentials, profile, or credential provider."
        )

    effective_api_key = (
        api_key
        if api_key is not None
        else current_api_key
        if api_key_was_mutated and token_provider is None and not aws_override
        else None
    )
    bearer_override = effective_api_key is not None or token_provider is not None

    routing_override = aws_region is not None or base_url is not None
    if not bearer_override and not aws_override and not routing_override:
        _refresh_legacy_provider_runtime(client)
        return {}, client._bedrock_provider, client._bedrock_state

    if bearer_override:
        next_api_key = effective_api_key
        next_token_provider = token_provider
        next_profile = None
        next_access_key_id = None
        next_secret_access_key = None
        next_session_token = None
        next_credentials_provider = None
    elif aws_override:
        next_api_key = None
        next_token_provider = None
        next_profile = aws_profile
        next_access_key_id = aws_access_key_id
        next_secret_access_key = aws_secret_access_key
        next_session_token = aws_session_token
        next_credentials_provider = aws_credentials_provider
    else:
        next_api_key = state.explicit_api_key
        next_token_provider = state.token_provider
        if state.uses_environment_bearer:
            next_api_key = state.environment_bearer_token or _environment_bearer_token()
            next_token_provider = None
        next_profile = state.aws_profile
        next_access_key_id = state.aws_access_key_id
        next_secret_access_key = state.aws_secret_access_key
        next_session_token = state.aws_session_token
        next_credentials_provider = state.aws_credentials_provider

    next_region = aws_region if aws_region is not None else client.aws_region
    next_region_was_explicit = aws_region is not None or state.region_was_explicit
    if aws_profile is not None and aws_region is None and not state.region_was_explicit:
        next_region = None

    if base_url is not None:
        next_base_url: str | httpx.URL | None = base_url
    elif state.uses_region_derived_base_url:
        next_base_url = ""
    else:
        next_base_url = client.base_url

    provider_kwargs: dict[str, object] = {
        "api_key": next_api_key,
        "bedrock_token_provider": next_token_provider,
        "aws_region": next_region,
        "aws_profile": next_profile,
        "aws_access_key_id": next_access_key_id,
        "aws_secret_access_key": next_secret_access_key,
        "aws_session_token": next_session_token,
        "aws_credentials_provider": next_credentials_provider,
        "base_url": next_base_url,
    }
    if _constructor_accepts_keyword(client.__class__.__init__, "_region_was_explicit"):
        provider_kwargs["_region_was_explicit"] = next_region_was_explicit

    return provider_kwargs, None, None


def _legacy_runtime_signature(
    client: BedrockOpenAI | AsyncBedrockOpenAI,
    configuration: _LegacyAuthConfiguration,
) -> _LegacyRuntimeSignature:
    mode, credential = configuration
    credential_identity: object = (
        hashlib.blake2s(credential.encode(), key=_LEGACY_SIGNATURE_KEY).digest()
        if isinstance(credential, str)
        else id(credential)
    )
    return _LegacyRuntimeSignature(
        mode=mode,
        base_url=str(client.base_url),
        region=client.aws_region,
        credential_identity=credential_identity,
    )


def _provider_for_legacy_client(
    client: BedrockOpenAI | AsyncBedrockOpenAI,
    configuration: _LegacyAuthConfiguration,
) -> _Provider:
    mode, credential = configuration
    if mode == "bearer":
        if not isinstance(credential, str) or not credential:
            raise OpenAIError("The Bedrock bearer credential must not be empty.")
        return bedrock(
            region=client.aws_region,
            base_url=client.base_url,
            api_key=credential,
        )
    if mode == "token_provider":
        return bedrock(
            region=client.aws_region,
            base_url=client.base_url,
            token_provider=cast("AsyncBedrockTokenProvider", credential),
        )

    state = client._bedrock_state
    return bedrock(
        region=client.aws_region,
        base_url=client.base_url,
        profile=state.aws_profile,
        access_key_id=state.aws_access_key_id,
        secret_access_key=state.aws_secret_access_key,
        session_token=state.aws_session_token,
        credential_provider=state.aws_credentials_provider,
    )


def _synchronize_legacy_routing_state(client: BedrockOpenAI | AsyncBedrockOpenAI) -> None:
    previous_signature = client._bedrock_runtime_signature
    base_url_changed = str(client.base_url) != previous_signature.base_url
    region_changed = client.aws_region != previous_signature.region
    if base_url_changed:
        client._bedrock_state = replace(client._bedrock_state, uses_region_derived_base_url=False)
        client._uses_region_derived_base_url = False
    if region_changed:
        client._bedrock_state = replace(
            client._bedrock_state,
            aws_region=client.aws_region,
            region_was_explicit=client.aws_region is not None,
        )
        if client._bedrock_state.uses_region_derived_base_url and client.aws_region is not None:
            client.base_url = f"https://bedrock-mantle.{client.aws_region}.api.aws/openai/v1"


def _refresh_legacy_provider_runtime(client: BedrockOpenAI | AsyncBedrockOpenAI) -> None:
    _synchronize_legacy_routing_state(client)
    configuration = client._legacy_auth_configuration()
    signature = _legacy_runtime_signature(client, configuration)
    if signature == client._bedrock_runtime_signature:
        return

    provider = _provider_for_legacy_client(client, configuration)
    client._bedrock_provider = provider
    client._provider = provider
    client._provider_runtime = _configure_provider(provider)
    if (
        isinstance(client._provider_runtime, _BedrockProviderRuntime)
        and client.aws_region is None
        and client._provider_runtime.region is not None
    ):
        client.aws_region = client._provider_runtime.region
        client._bedrock_state = replace(client._bedrock_state, aws_region=client.aws_region)
    client._bedrock_runtime_signature = _legacy_runtime_signature(client, configuration)


class BedrockOpenAI(OpenAI):
    """Compatibility client for Amazon Bedrock's OpenAI-compatible endpoint."""

    _bedrock_provider: _Provider
    _bedrock_state: _LegacyBedrockState
    _bedrock_token_provider: BedrockTokenProvider | None
    _uses_region_derived_base_url: bool
    _bedrock_runtime_signature: _LegacyRuntimeSignature
    aws_region: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        bedrock_token_provider: BedrockTokenProvider | None = None,
        aws_region: str | None = None,
        aws_profile: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_credentials_provider: AwsCredentialsProvider | None = None,
        organization: str | None = None,
        project: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        websocket_base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        _strict_response_validation: bool = False,
        _enforce_credentials: bool = True,
        _provider: _Provider | None = None,
        _state: _LegacyBedrockState | None = None,
        _region_was_explicit: bool | None = None,
    ) -> None:
        if _provider is None or _state is None:
            _provider, _state, public_api_key = _legacy_provider(
                api_key=api_key,
                token_provider=bedrock_token_provider,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                aws_credentials_provider=aws_credentials_provider,
                base_url=base_url,
                region_was_explicit=_region_was_explicit,
            )
        else:
            public_api_key = (
                _state.explicit_api_key
                or (_state.environment_bearer_token if _state.uses_environment_bearer else "")
                or ""
            )

        super().__init__(
            provider=_provider,
            organization=organization,
            project=project,
            webhook_secret=webhook_secret,
            websocket_base_url=websocket_base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
            _enforce_credentials=False,
        )

        self._bedrock_provider = _provider
        self._bedrock_state = _state
        self._bedrock_token_provider = cast("BedrockTokenProvider | None", _state.token_provider)
        self._uses_region_derived_base_url = _state.uses_region_derived_base_url
        canonical_region = re.fullmatch(r"bedrock-mantle\.([a-z0-9-]+)\.api\.aws", self.base_url.host)
        provider_region = (
            self._provider_runtime.region if isinstance(self._provider_runtime, _BedrockProviderRuntime) else None
        )
        self.aws_region = (
            _state.aws_region
            or provider_region
            or (canonical_region.group(1) if canonical_region is not None else None)
        )
        self._bedrock_state = replace(_state, aws_region=self.aws_region)
        self.api_key = public_api_key or ""
        self._bedrock_runtime_signature = _legacy_runtime_signature(self, self._legacy_auth_configuration())

    def _legacy_auth_configuration(self) -> _LegacyAuthConfiguration:
        if self._bedrock_token_provider is not None:
            return ("token_provider", self._bedrock_token_provider)
        if (
            self._bedrock_state.explicit_api_key is not None
            or self._bedrock_state.uses_environment_bearer
            or self.api_key
        ):
            return ("bearer", self.api_key)
        return ("aws", None)

    def _uses_aws_auth(self) -> bool:
        return (
            self._bedrock_state.explicit_api_key is None
            and not self.api_key
            and self._bedrock_token_provider is None
            and not self._bedrock_state.uses_environment_bearer
        )

    @override
    def _refresh_api_key(self) -> str:
        if self._bedrock_state.uses_environment_bearer:
            captured = self._bedrock_state.environment_bearer_token or ""
            return self.api_key if self.api_key and self.api_key != captured else captured
        if self._bedrock_token_provider is not None:
            token = cast(object, self._bedrock_token_provider())
            if not isinstance(token, str) or not token:
                raise ValueError("Expected `bedrock_token_provider` argument to return a non-empty string.")
            return token
        return self.api_key

    @override
    def _prepare_options(self, options: FinalRequestOptions) -> FinalRequestOptions:
        _refresh_legacy_provider_runtime(self)
        return super()._prepare_options(options)

    @override
    def copy(
        self,
        *,
        api_key: str | BedrockTokenProvider | None = None,
        admin_api_key: str | None = None,
        workload_identity: WorkloadIdentity | None = None,
        provider: _Provider | None | NotGiven = NOT_GIVEN,
        bedrock_token_provider: BedrockTokenProvider | None = None,
        aws_region: str | None = None,
        aws_profile: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_credentials_provider: AwsCredentialsProvider | None = None,
        organization: str | None = None,
        project: str | None = None,
        webhook_secret: str | None = None,
        websocket_base_url: str | httpx.URL | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _enforce_credentials: bool | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        if callable(api_key):
            raise OpenAIError("Pass refreshable Bedrock credentials via `bedrock_token_provider`, not `api_key`.")
        if not isinstance(provider, NotGiven):
            raise OpenAIError("Configure `provider` on `OpenAI`, not on `BedrockOpenAI.with_options()`.")
        if admin_api_key is not None or workload_identity is not None:
            raise OpenAIError("BedrockOpenAI only supports Bedrock bearer token or AWS credential authentication.")
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")
        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers
        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        provider_kwargs, inherited_provider, inherited_state = _copy_configuration(
            self,
            api_key=api_key,
            token_provider=bedrock_token_provider,
            aws_region=aws_region,
            aws_profile=aws_profile,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_credentials_provider=aws_credentials_provider,
            base_url=base_url,
        )
        constructor_kwargs: dict[str, Any] = {
            **provider_kwargs,
            "organization": organization if organization is not None else self.organization,
            "project": project if project is not None else self.project,
            "webhook_secret": webhook_secret if webhook_secret is not None else self.webhook_secret,
            "websocket_base_url": websocket_base_url if websocket_base_url is not None else self.websocket_base_url,
            "timeout": self.timeout if isinstance(timeout, NotGiven) else timeout,
            "http_client": http_client or self._client,
            "max_retries": max_retries if is_given(max_retries) else self.max_retries,
            "default_headers": headers,
            "default_query": params,
            "_enforce_credentials": True if _enforce_credentials is None else _enforce_credentials,
            **_extra_kwargs,
        }
        if inherited_provider is not None and _constructor_accepts_keyword(self.__class__.__init__, "_provider"):
            constructor_kwargs["_provider"] = inherited_provider
            constructor_kwargs["_state"] = inherited_state
        elif inherited_provider is not None:
            constructor_kwargs.update(
                api_key=self._bedrock_state.explicit_api_key or self._bedrock_state.environment_bearer_token,
                bedrock_token_provider=self._bedrock_state.token_provider,
                aws_region=self._bedrock_state.aws_region,
                aws_profile=self._bedrock_state.aws_profile,
                aws_access_key_id=self._bedrock_state.aws_access_key_id,
                aws_secret_access_key=self._bedrock_state.aws_secret_access_key,
                aws_session_token=self._bedrock_state.aws_session_token,
                aws_credentials_provider=self._bedrock_state.aws_credentials_provider,
                base_url="" if self._bedrock_state.uses_region_derived_base_url else self.base_url,
            )
            constructor_kwargs = {
                name: value
                for name, value in constructor_kwargs.items()
                if _constructor_accepts_keyword(self.__class__.__init__, name)
            }
        elif self.__class__ is not BedrockOpenAI:
            constructor_kwargs = {
                name: value
                for name, value in constructor_kwargs.items()
                if value is not None or _constructor_accepts_keyword(self.__class__.__init__, name)
            }
        return self.__class__(**constructor_kwargs)

    with_options = copy


class AsyncBedrockOpenAI(AsyncOpenAI):
    """Async compatibility client for Amazon Bedrock's OpenAI-compatible endpoint."""

    _bedrock_provider: _Provider
    _bedrock_state: _LegacyBedrockState
    _bedrock_token_provider: AsyncBedrockTokenProvider | None
    _uses_region_derived_base_url: bool
    _bedrock_runtime_signature: _LegacyRuntimeSignature
    aws_region: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        bedrock_token_provider: AsyncBedrockTokenProvider | None = None,
        aws_region: str | None = None,
        aws_profile: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_credentials_provider: AwsCredentialsProvider | None = None,
        organization: str | None = None,
        project: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        websocket_base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.AsyncClient | None = None,
        _strict_response_validation: bool = False,
        _enforce_credentials: bool = True,
        _provider: _Provider | None = None,
        _state: _LegacyBedrockState | None = None,
        _region_was_explicit: bool | None = None,
    ) -> None:
        if _provider is None or _state is None:
            _provider, _state, public_api_key = _legacy_provider(
                api_key=api_key,
                token_provider=bedrock_token_provider,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                aws_credentials_provider=aws_credentials_provider,
                base_url=base_url,
                region_was_explicit=_region_was_explicit,
            )
        else:
            public_api_key = (
                _state.explicit_api_key
                or (_state.environment_bearer_token if _state.uses_environment_bearer else "")
                or ""
            )

        super().__init__(
            provider=_provider,
            organization=organization,
            project=project,
            webhook_secret=webhook_secret,
            websocket_base_url=websocket_base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
            _enforce_credentials=False,
        )

        self._bedrock_provider = _provider
        self._bedrock_state = _state
        self._bedrock_token_provider = cast("AsyncBedrockTokenProvider | None", _state.token_provider)
        self._uses_region_derived_base_url = _state.uses_region_derived_base_url
        canonical_region = re.fullmatch(r"bedrock-mantle\.([a-z0-9-]+)\.api\.aws", self.base_url.host)
        provider_region = (
            self._provider_runtime.region if isinstance(self._provider_runtime, _BedrockProviderRuntime) else None
        )
        self.aws_region = (
            _state.aws_region
            or provider_region
            or (canonical_region.group(1) if canonical_region is not None else None)
        )
        self._bedrock_state = replace(_state, aws_region=self.aws_region)
        self.api_key = public_api_key or ""
        self._bedrock_runtime_signature = _legacy_runtime_signature(self, self._legacy_auth_configuration())

    def _legacy_auth_configuration(self) -> _LegacyAuthConfiguration:
        if self._bedrock_token_provider is not None:
            return ("token_provider", self._bedrock_token_provider)
        if (
            self._bedrock_state.explicit_api_key is not None
            or self._bedrock_state.uses_environment_bearer
            or self.api_key
        ):
            return ("bearer", self.api_key)
        return ("aws", None)

    def _uses_aws_auth(self) -> bool:
        return (
            self._bedrock_state.explicit_api_key is None
            and not self.api_key
            and self._bedrock_token_provider is None
            and not self._bedrock_state.uses_environment_bearer
        )

    @override
    async def _refresh_api_key(self) -> str:
        if self._bedrock_state.uses_environment_bearer:
            captured = self._bedrock_state.environment_bearer_token or ""
            return self.api_key if self.api_key and self.api_key != captured else captured
        if self._bedrock_token_provider is not None:
            token = cast(object, self._bedrock_token_provider())
            if inspect.isawaitable(token):
                token = await token
            if not isinstance(token, str) or not token:
                raise ValueError("Expected `bedrock_token_provider` argument to return a non-empty string.")
            return token
        return self.api_key

    @override
    async def _prepare_options(self, options: FinalRequestOptions) -> FinalRequestOptions:
        _refresh_legacy_provider_runtime(self)
        return await super()._prepare_options(options)

    @override
    def copy(
        self,
        *,
        api_key: str | AsyncBedrockTokenProvider | None = None,
        admin_api_key: str | None = None,
        workload_identity: WorkloadIdentity | None = None,
        provider: _Provider | None | NotGiven = NOT_GIVEN,
        bedrock_token_provider: AsyncBedrockTokenProvider | None = None,
        aws_region: str | None = None,
        aws_profile: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_credentials_provider: AwsCredentialsProvider | None = None,
        organization: str | None = None,
        project: str | None = None,
        webhook_secret: str | None = None,
        websocket_base_url: str | httpx.URL | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _enforce_credentials: bool | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        if callable(api_key):
            raise OpenAIError("Pass refreshable Bedrock credentials via `bedrock_token_provider`, not `api_key`.")
        if not isinstance(provider, NotGiven):
            raise OpenAIError("Configure `provider` on `AsyncOpenAI`, not on `AsyncBedrockOpenAI.with_options()`.")
        if admin_api_key is not None or workload_identity is not None:
            raise OpenAIError("AsyncBedrockOpenAI only supports Bedrock bearer token or AWS credential authentication.")
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")
        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers
        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        provider_kwargs, inherited_provider, inherited_state = _copy_configuration(
            self,
            api_key=api_key,
            token_provider=bedrock_token_provider,
            aws_region=aws_region,
            aws_profile=aws_profile,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            aws_credentials_provider=aws_credentials_provider,
            base_url=base_url,
        )
        constructor_kwargs: dict[str, Any] = {
            **provider_kwargs,
            "organization": organization if organization is not None else self.organization,
            "project": project if project is not None else self.project,
            "webhook_secret": webhook_secret if webhook_secret is not None else self.webhook_secret,
            "websocket_base_url": websocket_base_url if websocket_base_url is not None else self.websocket_base_url,
            "timeout": self.timeout if isinstance(timeout, NotGiven) else timeout,
            "http_client": http_client or self._client,
            "max_retries": max_retries if is_given(max_retries) else self.max_retries,
            "default_headers": headers,
            "default_query": params,
            "_enforce_credentials": True if _enforce_credentials is None else _enforce_credentials,
            **_extra_kwargs,
        }
        if inherited_provider is not None and _constructor_accepts_keyword(self.__class__.__init__, "_provider"):
            constructor_kwargs["_provider"] = inherited_provider
            constructor_kwargs["_state"] = inherited_state
        elif inherited_provider is not None:
            constructor_kwargs.update(
                api_key=self._bedrock_state.explicit_api_key or self._bedrock_state.environment_bearer_token,
                bedrock_token_provider=self._bedrock_state.token_provider,
                aws_region=self._bedrock_state.aws_region,
                aws_profile=self._bedrock_state.aws_profile,
                aws_access_key_id=self._bedrock_state.aws_access_key_id,
                aws_secret_access_key=self._bedrock_state.aws_secret_access_key,
                aws_session_token=self._bedrock_state.aws_session_token,
                aws_credentials_provider=self._bedrock_state.aws_credentials_provider,
                base_url="" if self._bedrock_state.uses_region_derived_base_url else self.base_url,
            )
            constructor_kwargs = {
                name: value
                for name, value in constructor_kwargs.items()
                if _constructor_accepts_keyword(self.__class__.__init__, name)
            }
        elif self.__class__ is not AsyncBedrockOpenAI:
            constructor_kwargs = {
                name: value
                for name, value in constructor_kwargs.items()
                if value is not None or _constructor_accepts_keyword(self.__class__.__init__, name)
            }
        return self.__class__(**constructor_kwargs)

    with_options = copy


__all__ = [
    "BedrockOpenAI",
    "AsyncBedrockOpenAI",
    "BedrockTokenProvider",
    "AsyncBedrockTokenProvider",
    "AwsCredentialsProvider",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["SkillVersion"]


class SkillVersion(BaseModel):
    id: str
    """Unique identifier for the skill version."""

    created_at: int
    """Unix timestamp (seconds) for when the version was created."""

    description: str
    """Description of the skill version."""

    name: str
    """Name of the skill version."""

    object: Literal["skill.version"]
    """The object type, which is `skill.version`."""

    skill_id: str
    """Identifier of the skill for this version."""

    version: str
    """Version number for this skill."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union, Optional
from typing_extensions import Literal, Annotated, TypeAlias

from ..._utils import PropertyInfo
from ..._models import BaseModel

__all__ = ["RealtimeAudioInputTurnDetection", "ServerVad", "SemanticVad"]


class ServerVad(BaseModel):
    """
    Server-side voice activity detection (VAD) which flips on when user speech is detected and off after a period of silence.
    """

    type: Literal["server_vad"]
    """Type of turn detection, `server_vad` to turn on simple Server VAD."""

    create_response: Optional[bool] = None
    """Whether or not to automatically generate a response when a VAD stop event
    occurs.

    If `interrupt_response` is set to `false` this may fail to create a response if
    the model is already responding.

    If both `create_response` and `interrupt_response` are set to `false`, the model
    will never respond automatically but VAD events will still be emitted.
    """

    idle_timeout_ms: Optional[int] = None
    """Optional timeout after which a model response will be triggered automatically.

    This is useful for situations in which a long pause from the user is unexpected,
    such as a phone call. The model will effectively prompt the user to continue the
    conversation based on the current context.

    The timeout value will be applied after the last model response's audio has
    finished playing, i.e. it's set to the `response.done` time plus audio playback
    duration.

    An `input_audio_buffer.timeout_triggered` event (plus events associated with the
    Response) will be emitted when the timeout is reached. Idle timeout is currently
    only supported for `server_vad` mode.
    """

    interrupt_response: Optional[bool] = None
    """
    Whether or not to automatically interrupt (cancel) any ongoing response with
    output to the default conversation (i.e. `conversation` of `auto`) when a VAD
    start event occurs. If `true` then the response will be cancelled, otherwise it
    will continue until complete.

    If both `create_response` and `interrupt_response` are set to `false`, the model
    will never respond automatically but VAD events will still be emitted.
    """

    prefix_padding_ms: Optional[int] = None
    """Used only for `server_vad` mode.

    Amount of audio to include before the VAD detected speech (in milliseconds).
    Defaults to 300ms.
    """

    silence_duration_ms: Optional[int] = None
    """Used only for `server_vad` mode.

    Duration of silence to detect speech stop (in milliseconds). Defaults to 500ms.
    With shorter values the model will respond more quickly, but may jump in on
    short pauses from the user.
    """

    threshold: Optional[float] = None
    """Used only for `server_vad` mode.

    Activation threshold for VAD (0.0 to 1.0), this defaults to 0.5. A higher
    threshold will require louder audio to activate the model, and thus might
    perform better in noisy environments.
    """


class SemanticVad(BaseModel):
    """
    Server-side semantic turn detection which uses a model to determine when the user has finished speaking.
    """

    type: Literal["semantic_vad"]
    """Type of turn detection, `semantic_vad` to turn on Semantic VAD."""

    create_response: Optional[bool] = None
    """
    Whether or not to automatically generate a response when a VAD stop event
    occurs.
    """

    eagerness: Optional[Literal["low", "medium", "high", "auto"]] = None
    """Used only for `semantic_vad` mode.

    The eagerness of the model to respond. `low` will wait longer for the user to
    continue speaking, `high` will respond more quickly. `auto` is the default and
    is equivalent to `medium`. `low`, `medium`, and `high` have max timeouts of 8s,
    4s, and 2s respectively.
    """

    interrupt_response: Optional[bool] = None
    """
    Whether or not to automatically interrupt any ongoing response with output to
    the default conversation (i.e. `conversation` of `auto`) when a VAD start event
    occurs.
    """


RealtimeAudioInputTurnDetection: TypeAlias = Annotated[
    Union[ServerVad, SemanticVad, None], PropertyInfo(discriminator="type")
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["BetaLocalSkillParam"]


class BetaLocalSkillParam(TypedDict, total=False):
    description: Required[str]
    """The description of the skill."""

    name: Required[str]
    """The name of the skill."""

    path: Required[str]
    """The path to the directory containing the skill."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from openai import OpenAI, AsyncOpenAI
from tests.utils import assert_matches_type
from openai.types.fine_tuning.alpha import (
    GraderRunResponse,
    GraderValidateResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestGraders:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_run(self, client: OpenAI) -> None:
        grader = client.fine_tuning.alpha.graders.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        )
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    def test_method_run_with_all_params(self, client: OpenAI) -> None:
        grader = client.fine_tuning.alpha.graders.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
            item={},
        )
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    def test_raw_response_run(self, client: OpenAI) -> None:
        response = client.fine_tuning.alpha.graders.with_raw_response.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        grader = response.parse()
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    def test_streaming_response_run(self, client: OpenAI) -> None:
        with client.fine_tuning.alpha.graders.with_streaming_response.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            grader = response.parse()
            assert_matches_type(GraderRunResponse, grader, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_validate(self, client: OpenAI) -> None:
        grader = client.fine_tuning.alpha.graders.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    def test_method_validate_with_all_params(self, client: OpenAI) -> None:
        grader = client.fine_tuning.alpha.graders.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    def test_raw_response_validate(self, client: OpenAI) -> None:
        response = client.fine_tuning.alpha.graders.with_raw_response.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        grader = response.parse()
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    def test_streaming_response_validate(self, client: OpenAI) -> None:
        with client.fine_tuning.alpha.graders.with_streaming_response.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            grader = response.parse()
            assert_matches_type(GraderValidateResponse, grader, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncGraders:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_run(self, async_client: AsyncOpenAI) -> None:
        grader = await async_client.fine_tuning.alpha.graders.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        )
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    async def test_method_run_with_all_params(self, async_client: AsyncOpenAI) -> None:
        grader = await async_client.fine_tuning.alpha.graders.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
            item={},
        )
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    async def test_raw_response_run(self, async_client: AsyncOpenAI) -> None:
        response = await async_client.fine_tuning.alpha.graders.with_raw_response.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        grader = response.parse()
        assert_matches_type(GraderRunResponse, grader, path=["response"])

    @parametrize
    async def test_streaming_response_run(self, async_client: AsyncOpenAI) -> None:
        async with async_client.fine_tuning.alpha.graders.with_streaming_response.run(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
            model_sample="model_sample",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            grader = await response.parse()
            assert_matches_type(GraderRunResponse, grader, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_validate(self, async_client: AsyncOpenAI) -> None:
        grader = await async_client.fine_tuning.alpha.graders.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    async def test_method_validate_with_all_params(self, async_client: AsyncOpenAI) -> None:
        grader = await async_client.fine_tuning.alpha.graders.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    async def test_raw_response_validate(self, async_client: AsyncOpenAI) -> None:
        response = await async_client.fine_tuning.alpha.graders.with_raw_response.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        grader = response.parse()
        assert_matches_type(GraderValidateResponse, grader, path=["response"])

    @parametrize
    async def test_streaming_response_validate(self, async_client: AsyncOpenAI) -> None:
        async with async_client.fine_tuning.alpha.graders.with_streaming_response.validate(
            grader={
                "input": "input",
                "name": "name",
                "operation": "eq",
                "reference": "reference",
                "type": "string_check",
            },
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            grader = await response.parse()
            assert_matches_type(GraderValidateResponse, grader, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Literal, Annotated, TypeAlias

from ..._utils import PropertyInfo
from ..._models import BaseModel
from .beta_response_format_text_json_schema_config import BetaResponseFormatTextJSONSchemaConfig

__all__ = ["BetaResponseFormatTextConfig", "Text", "JSONObject"]


class Text(BaseModel):
    """Default response format. Used to generate text responses."""

    type: Literal["text"]
    """The type of response format being defined. Always `text`."""


class JSONObject(BaseModel):
    """JSON object response format.

    An older method of generating JSON responses.
    Using `json_schema` is recommended for models that support it. Note that the
    model will not generate JSON without a system or user message instructing it
    to do so.
    """

    type: Literal["json_object"]
    """The type of response format being defined. Always `json_object`."""


BetaResponseFormatTextConfig: TypeAlias = Annotated[
    Union[Text, BetaResponseFormatTextJSONSchemaConfig, JSONObject], PropertyInfo(discriminator="type")
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Literal, TypeAlias

__all__ = ["VideoModel"]

VideoModel: TypeAlias = Union[
    str, Literal["sora-2", "sora-2-pro", "sora-2-2025-10-06", "sora-2-pro-2025-10-06", "sora-2-2025-12-08"]
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["Role"]


class Role(BaseModel):
    """Details about a role that can be assigned through the public Roles API."""

    id: str
    """Identifier for the role."""

    description: Optional[str] = None
    """Optional description of the role."""

    name: str
    """Unique name for the role."""

    object: Literal["role"]
    """Always `role`."""

    permissions: List[str]
    """Permissions granted by the role."""

    predefined_role: bool
    """Whether the role is predefined and managed by OpenAI."""

    resource_type: str
    """
    Resource type the role is bound to (for example `api.organization` or
    `api.project`).
    """



from __future__ import annotations

import sys
import email
import tarfile
import zipfile
import argparse
import platform
import tempfile
import subprocess
from pathlib import Path
from email.message import Message

ROOT = Path(__file__).resolve().parents[2]
EXPECTED_REQUIRES_PYTHON = ">=3.10"


def wheel_metadata(wheel: Path) -> Message:
    with zipfile.ZipFile(wheel) as archive:
        names = [name for name in archive.namelist() if name.endswith(".dist-info/METADATA")]
        if len(names) != 1:
            raise RuntimeError(f"Expected exactly one METADATA file in {wheel}, found: {names}")
        return email.message_from_bytes(archive.read(names[0]))


def sdist_metadata(sdist: Path) -> Message:
    with tarfile.open(sdist) as archive:
        members = [member for member in archive.getmembers() if member.name.endswith("/PKG-INFO")]
        if len(members) != 1:
            raise RuntimeError(f"Expected exactly one PKG-INFO file in {sdist}, found: {members}")
        extracted = archive.extractfile(members[0])
        if extracted is None:
            raise RuntimeError(f"Could not read {members[0].name} from {sdist}")
        return email.message_from_bytes(extracted.read())


def assert_python_39_rejected(wheel: Path) -> None:
    if platform.python_version_tuple()[:2] != ("3", "9"):
        raise RuntimeError(f"The resolver check must run on Python 3.9, found {sys.version.split()[0]}")

    with tempfile.TemporaryDirectory(prefix="openai-python-version-wheel-") as directory:
        environment_path = Path(directory) / "venv"
        subprocess.run([sys.executable, "-m", "venv", str(environment_path)], check=True)
        python = environment_path / ("Scripts/python.exe" if sys.platform == "win32" else "bin/python")
        result = subprocess.run(
            [
                str(python),
                "-m",
                "pip",
                "install",
                "--no-input",
                "--ignore-installed",
                "--no-deps",
                str(wheel.resolve()),
            ],
            cwd=ROOT,
            capture_output=True,
            text=True,
            check=False,
        )
    if result.returncode == 0:
        raise RuntimeError("Python 3.9 unexpectedly accepted the built wheel")

    output = result.stdout + result.stderr
    if "3.10" not in output or "Python" not in output:
        raise RuntimeError(f"Python 3.9 rejection failed for an unexpected reason:\n{output}")


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--check-python-39",
        action="store_true",
        help="Also prove that pip on an actual Python 3.9 interpreter rejects the wheel",
    )
    args = parser.parse_args()

    wheels = list((ROOT / "dist").glob("*.whl"))
    sdists = list((ROOT / "dist").glob("*.tar.gz"))
    if len(wheels) != 1 or len(sdists) != 1:
        raise RuntimeError(f"Expected one wheel and one sdist, found wheels={wheels}, sdists={sdists}")

    for artifact, metadata in (
        (wheels[0], wheel_metadata(wheels[0])),
        (sdists[0], sdist_metadata(sdists[0])),
    ):
        actual = metadata["Requires-Python"]
        if actual != EXPECTED_REQUIRES_PYTHON:
            raise RuntimeError(f"Expected {artifact.name} to require {EXPECTED_REQUIRES_PYTHON}, found {actual}")

    if args.check_python_39:
        assert_python_39_rejected(wheels[0])
        print("Validated wheel and sdist Python metadata and Python 3.9 rejection")
    else:
        print("Validated wheel and sdist Python metadata")


if __name__ == "__main__":
    main()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["CertificateCreateParams"]


class CertificateCreateParams(TypedDict, total=False):
    certificate: Required[str]
    """The certificate content in PEM format"""

    name: str
    """An optional name for the certificate"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ResponseCompactionItem"]


class ResponseCompactionItem(BaseModel):
    """
    A compaction item generated by the [`v1/responses/compact` API](https://platform.openai.com/docs/api-reference/responses/compact).
    """

    id: str
    """The unique ID of the compaction item."""

    encrypted_content: str
    """The encrypted content that was produced by compaction."""

    type: Literal["compaction"]
    """The type of the item. Always `compaction`."""

    created_by: Optional[str] = None
    """The identifier of the actor that created the item."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel
from .realtime_reasoning_effort import RealtimeReasoningEffort

__all__ = ["RealtimeReasoning"]


class RealtimeReasoning(BaseModel):
    """Configuration for reasoning-capable Realtime models such as `gpt-realtime-2`."""

    effort: Optional[RealtimeReasoningEffort] = None
    """
    Constrains effort on reasoning for reasoning-capable Realtime models such as
    `gpt-realtime-2`.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import Literal, Required, TypeAlias, TypedDict

__all__ = ["BetaResponseFunctionShellCallOutputContentParam", "Outcome", "OutcomeTimeout", "OutcomeExit"]


class OutcomeTimeout(TypedDict, total=False):
    """Indicates that the shell call exceeded its configured time limit."""

    type: Required[Literal["timeout"]]
    """The outcome type. Always `timeout`."""


class OutcomeExit(TypedDict, total=False):
    """Indicates that the shell commands finished and returned an exit code."""

    exit_code: Required[int]
    """The exit code returned by the shell process."""

    type: Required[Literal["exit"]]
    """The outcome type. Always `exit`."""


Outcome: TypeAlias = Union[OutcomeTimeout, OutcomeExit]


class BetaResponseFunctionShellCallOutputContentParam(TypedDict, total=False):
    """Captured stdout and stderr for a portion of a shell tool call output."""

    outcome: Required[Outcome]
    """The exit or timeout outcome associated with this shell call."""

    stderr: Required[str]
    """Captured stderr output for the shell call."""

    stdout: Required[str]
    """Captured stdout output for the shell call."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Union, Optional
from typing_extensions import Literal, Required, TypedDict

from .realtime_reasoning_param import RealtimeReasoningParam
from .realtime_truncation_param import RealtimeTruncationParam
from .realtime_audio_config_param import RealtimeAudioConfigParam
from .realtime_tools_config_param import RealtimeToolsConfigParam
from .realtime_tracing_config_param import RealtimeTracingConfigParam
from ..responses.response_prompt_param import ResponsePromptParam
from .realtime_tool_choice_config_param import RealtimeToolChoiceConfigParam

__all__ = ["RealtimeSessionCreateRequestParam"]


class RealtimeSessionCreateRequestParam(TypedDict, total=False):
    """Realtime session object configuration."""

    type: Required[Literal["realtime"]]
    """The type of session to create. Always `realtime` for the Realtime API."""

    audio: RealtimeAudioConfigParam
    """Configuration for input and output audio."""

    include: List[Literal["item.input_audio_transcription.logprobs"]]
    """Additional fields to include in server outputs.

    `item.input_audio_transcription.logprobs`: Include logprobs for input audio
    transcription.
    """

    instructions: str
    """The default system instructions (i.e.

    system message) prepended to model calls. This field allows the client to guide
    the model on desired responses. The model can be instructed on response content
    and format, (e.g. "be extremely succinct", "act friendly", "here are examples of
    good responses") and on audio behavior (e.g. "talk quickly", "inject emotion
    into your voice", "laugh frequently"). The instructions are not guaranteed to be
    followed by the model, but they provide guidance to the model on the desired
    behavior.

    Note that the server sets default instructions which will be used if this field
    is not set and are visible in the `session.created` event at the start of the
    session.
    """

    max_output_tokens: Union[int, Literal["inf"]]
    """
    Maximum number of output tokens for a single assistant response, inclusive of
    tool calls. Provide an integer between 1 and 4096 to limit output tokens, or
    `inf` for the maximum available tokens for a given model. Defaults to `inf`.
    """

    model: Union[
        str,
        Literal[
            "gpt-realtime",
            "gpt-realtime-1.5",
            "gpt-realtime-2",
            "gpt-realtime-2.1",
            "gpt-realtime-2.1-mini",
            "gpt-realtime-2025-08-28",
            "gpt-4o-realtime-preview",
            "gpt-4o-realtime-preview-2024-10-01",
            "gpt-4o-realtime-preview-2024-12-17",
            "gpt-4o-realtime-preview-2025-06-03",
            "gpt-4o-mini-realtime-preview",
            "gpt-4o-mini-realtime-preview-2024-12-17",
            "gpt-realtime-mini",
            "gpt-realtime-mini-2025-10-06",
            "gpt-realtime-mini-2025-12-15",
            "gpt-audio-1.5",
            "gpt-audio-mini",
            "gpt-audio-mini-2025-10-06",
            "gpt-audio-mini-2025-12-15",
        ],
    ]
    """The Realtime model used for this session."""

    output_modalities: List[Literal["text", "audio"]]
    """The set of modalities the model can respond with.

    It defaults to `["audio"]`, indicating that the model will respond with audio
    plus a transcript. `["text"]` can be used to make the model respond with text
    only. It is not possible to request both `text` and `audio` at the same time.
    """

    parallel_tool_calls: bool
    """Whether the model may call multiple tools in parallel.

    Only supported by reasoning Realtime models such as `gpt-realtime-2`.
    """

    prompt: Optional[ResponsePromptParam]
    """
    Reference to a prompt template and its variables.
    [Learn more](https://platform.openai.com/docs/guides/text?api-mode=responses#reusable-prompts).
    """

    reasoning: RealtimeReasoningParam
    """Configuration for reasoning-capable Realtime models such as `gpt-realtime-2`."""

    tool_choice: RealtimeToolChoiceConfigParam
    """How the model chooses tools.

    Provide one of the string modes or force a specific function/MCP tool.
    """

    tools: RealtimeToolsConfigParam
    """Tools available to the model."""

    tracing: Optional[RealtimeTracingConfigParam]
    """
    Realtime API can write session traces to the
    [Traces Dashboard](https://platform.openai.com/logs?api=traces). Set to null to
    disable tracing. Once tracing is enabled for a session, the configuration cannot
    be modified.

    `auto` will create a trace for the session with default values for the workflow
    name, group id, and metadata.
    """

    truncation: RealtimeTruncationParam
    """
    When the number of tokens in a conversation exceeds the model's input token
    limit, the conversation be truncated, meaning messages (starting from the
    oldest) will not be included in the model's context. A 32k context model with
    4,096 max output tokens can only include 28,224 tokens in the context before
    truncation occurs.

    Clients can configure truncation behavior to truncate with a lower max token
    limit, which is an effective way to control token usage and cost.

    Truncation will reduce the number of cached tokens on the next turn (busting the
    cache), since messages are dropped from the beginning of the context. However,
    clients can also configure truncation to retain messages up to a fraction of the
    maximum context size, which will reduce the need for future truncations and thus
    improve the cache rate.

    Truncation can be disabled entirely, which means the server will never truncate
    but would instead return an error if the conversation exceeds the model's input
    token limit.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .transcription_language import TranscriptionLanguage

__all__ = ["TranscriptionTextDoneEvent", "Logprob", "Usage", "UsageInputTokenDetails"]


class Logprob(BaseModel):
    token: Optional[str] = None
    """The token that was used to generate the log probability."""

    bytes: Optional[List[int]] = None
    """The bytes that were used to generate the log probability."""

    logprob: Optional[float] = None
    """The log probability of the token."""


class UsageInputTokenDetails(BaseModel):
    """Details about the input tokens billed for this request."""

    audio_tokens: Optional[int] = None
    """Number of audio tokens billed for this request."""

    text_tokens: Optional[int] = None
    """Number of text tokens billed for this request."""


class Usage(BaseModel):
    """Usage statistics for models billed by token usage."""

    input_tokens: int
    """Number of input tokens billed for this request."""

    output_tokens: int
    """Number of output tokens generated."""

    total_tokens: int
    """Total number of tokens used (input + output)."""

    type: Literal["tokens"]
    """The type of the usage object. Always `tokens` for this variant."""

    input_token_details: Optional[UsageInputTokenDetails] = None
    """Details about the input tokens billed for this request."""


class TranscriptionTextDoneEvent(BaseModel):
    """Emitted when the transcription is complete.

    Contains the complete transcription text. Only emitted when you [create a transcription](https://platform.openai.com/docs/api-reference/audio/create-transcription) with the `Stream` parameter set to `true`.
    """

    text: str
    """The text that was transcribed."""

    type: Literal["transcript.text.done"]
    """The type of the event. Always `transcript.text.done`."""

    languages: Optional[List[TranscriptionLanguage]] = None
    """The languages detected in the audio.

    Returned by `gpt-transcribe`. An empty array indicates that no language could be
    reliably detected.
    """

    logprobs: Optional[List[Logprob]] = None
    """The log probabilities of the individual tokens in the transcription.

    Only included if you
    [create a transcription](https://platform.openai.com/docs/api-reference/audio/create-transcription)
    with the `include[]` parameter set to `logprobs`.
    """

    usage: Optional[Usage] = None
    """Usage statistics for models billed by token usage."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .log_prob_properties import LogProbProperties

__all__ = ["ConversationItemInputAudioTranscriptionDeltaEvent"]


class ConversationItemInputAudioTranscriptionDeltaEvent(BaseModel):
    """
    Returned when the text value of an input audio transcription content part is updated with incremental transcription results.
    """

    event_id: str
    """The unique ID of the server event."""

    item_id: str
    """The ID of the item containing the audio that is being transcribed."""

    type: Literal["conversation.item.input_audio_transcription.delta"]
    """The event type, must be `conversation.item.input_audio_transcription.delta`."""

    content_index: Optional[int] = None
    """The index of the content part in the item's content array."""

    delta: Optional[str] = None
    """The text delta."""

    logprobs: Optional[List[LogProbProperties]] = None
    """The log probabilities of the transcription.

    These can be enabled by configurating the session with
    `"include": ["item.input_audio_transcription.logprobs"]`. Each entry in the
    array corresponds a log probability of which token would be selected for this
    chunk of transcription. This can help to identify if it was possible there were
    multiple valid options for a given chunk of transcription.
    """



from typing import List

from openai import OpenAI
from openai.types.responses.tool_param import ToolParam
from openai.types.responses.response_input_item_param import ResponseInputItemParam


def main() -> None:
    client = OpenAI()
    tools: List[ToolParam] = [
        {
            "type": "function",
            "name": "get_current_weather",
            "description": "Get current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "City and state, e.g. San Francisco, CA",
                    },
                    "unit": {
                        "type": "string",
                        "enum": ["c", "f"],
                        "description": "Temperature unit to use",
                    },
                },
                "required": ["location", "unit"],
                "additionalProperties": False,
            },
            "strict": True,
        }
    ]

    input_items: List[ResponseInputItemParam] = [
        {
            "type": "message",
            "role": "user",
            "content": [{"type": "input_text", "text": "What's the weather in San Francisco today?"}],
        }
    ]

    response = client.responses.input_tokens.count(
        model="gpt-5",
        instructions="You are a concise assistant.",
        input=input_items,
        tools=tools,
        tool_choice={"type": "function", "name": "get_current_weather"},
    )
    print(f"input tokens: {response.input_tokens}")


if __name__ == "__main__":
    main()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["RealtimeConnectParams"]


class RealtimeConnectParams(TypedDict, total=False):
    call_id: str

    model: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

from .image_file_param import ImageFileParam

__all__ = ["ImageFileContentBlockParam"]


class ImageFileContentBlockParam(TypedDict, total=False):
    """
    References an image [File](https://platform.openai.com/docs/api-reference/files) in the content of a message.
    """

    image_file: Required[ImageFileParam]

    type: Required[Literal["image_file"]]
    """Always `image_file`."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Mapping, cast
from typing_extensions import Literal

import httpx

from .... import _legacy_response
from .content import (
    Content,
    AsyncContent,
    ContentWithRawResponse,
    AsyncContentWithRawResponse,
    ContentWithStreamingResponse,
    AsyncContentWithStreamingResponse,
)
from ...._files import deepcopy_with_paths
from ...._types import (
    Body,
    Omit,
    Query,
    Headers,
    NotGiven,
    FileTypes,
    SequenceNotStr,
    omit,
    not_given,
)
from ...._utils import extract_files, path_template, maybe_transform, async_maybe_transform
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ....pagination import SyncCursorPage, AsyncCursorPage
from ...._base_client import AsyncPaginator, make_request_options
from ....types.skills import version_list_params, version_create_params
from ....types.skills.skill_version import SkillVersion
from ....types.skills.deleted_skill_version import DeletedSkillVersion

__all__ = ["Versions", "AsyncVersions"]


class Versions(SyncAPIResource):
    @cached_property
    def content(self) -> Content:
        return Content(self._client)

    @cached_property
    def with_raw_response(self) -> VersionsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return VersionsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> VersionsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return VersionsWithStreamingResponse(self)

    def create(
        self,
        skill_id: str,
        *,
        default: bool | Omit = omit,
        files: Union[SequenceNotStr[FileTypes], FileTypes] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SkillVersion:
        """
        Create a new immutable skill version.

        Args:
          default: Whether to set this version as the default.

          files: Skill files to upload (directory upload) or a single zip file.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        body = deepcopy_with_paths(
            {
                "default": default,
                "files": files,
            },
            [["files", "<array>"], ["files"]],
        )
        extracted_files = extract_files(cast(Mapping[str, object], body), paths=[["files", "<array>"], ["files"]])
        if extracted_files:
            # It should be noted that the actual Content-Type header that will be
            # sent to the server will contain a `boundary` parameter, e.g.
            # multipart/form-data; boundary=---abc--
            extra_headers = {"Content-Type": "multipart/form-data", **(extra_headers or {})}
        return self._post(
            path_template("/skills/{skill_id}/versions", skill_id=skill_id),
            body=maybe_transform(body, version_create_params.VersionCreateParams),
            files=extracted_files,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=SkillVersion,
        )

    def retrieve(
        self,
        version: str,
        *,
        skill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SkillVersion:
        """
        Get a specific skill version.

        Args:
          version: The version number to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return self._get(
            path_template("/skills/{skill_id}/versions/{version}", skill_id=skill_id, version=version),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=SkillVersion,
        )

    def list(
        self,
        skill_id: str,
        *,
        after: str | Omit = omit,
        limit: int | Omit = omit,
        order: Literal["asc", "desc"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncCursorPage[SkillVersion]:
        """
        List skill versions for a skill.

        Args:
          after: The skill version ID to start after.

          limit: Number of versions to retrieve.

          order: Sort order of results by version number.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        return self._get_api_list(
            path_template("/skills/{skill_id}/versions", skill_id=skill_id),
            page=SyncCursorPage[SkillVersion],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                        "order": order,
                    },
                    version_list_params.VersionListParams,
                ),
                security={"bearer_auth": True},
            ),
            model=SkillVersion,
        )

    def delete(
        self,
        version: str,
        *,
        skill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> DeletedSkillVersion:
        """
        Delete a skill version.

        Args:
          version: The skill version number.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return self._delete(
            path_template("/skills/{skill_id}/versions/{version}", skill_id=skill_id, version=version),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=DeletedSkillVersion,
        )


class AsyncVersions(AsyncAPIResource):
    @cached_property
    def content(self) -> AsyncContent:
        return AsyncContent(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncVersionsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return AsyncVersionsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncVersionsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return AsyncVersionsWithStreamingResponse(self)

    async def create(
        self,
        skill_id: str,
        *,
        default: bool | Omit = omit,
        files: Union[SequenceNotStr[FileTypes], FileTypes] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SkillVersion:
        """
        Create a new immutable skill version.

        Args:
          default: Whether to set this version as the default.

          files: Skill files to upload (directory upload) or a single zip file.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        body = deepcopy_with_paths(
            {
                "default": default,
                "files": files,
            },
            [["files", "<array>"], ["files"]],
        )
        extracted_files = extract_files(cast(Mapping[str, object], body), paths=[["files", "<array>"], ["files"]])
        if extracted_files:
            # It should be noted that the actual Content-Type header that will be
            # sent to the server will contain a `boundary` parameter, e.g.
            # multipart/form-data; boundary=---abc--
            extra_headers = {"Content-Type": "multipart/form-data", **(extra_headers or {})}
        return await self._post(
            path_template("/skills/{skill_id}/versions", skill_id=skill_id),
            body=await async_maybe_transform(body, version_create_params.VersionCreateParams),
            files=extracted_files,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=SkillVersion,
        )

    async def retrieve(
        self,
        version: str,
        *,
        skill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SkillVersion:
        """
        Get a specific skill version.

        Args:
          version: The version number to retrieve.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return await self._get(
            path_template("/skills/{skill_id}/versions/{version}", skill_id=skill_id, version=version),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=SkillVersion,
        )

    def list(
        self,
        skill_id: str,
        *,
        after: str | Omit = omit,
        limit: int | Omit = omit,
        order: Literal["asc", "desc"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[SkillVersion, AsyncCursorPage[SkillVersion]]:
        """
        List skill versions for a skill.

        Args:
          after: The skill version ID to start after.

          limit: Number of versions to retrieve.

          order: Sort order of results by version number.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        return self._get_api_list(
            path_template("/skills/{skill_id}/versions", skill_id=skill_id),
            page=AsyncCursorPage[SkillVersion],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                        "order": order,
                    },
                    version_list_params.VersionListParams,
                ),
                security={"bearer_auth": True},
            ),
            model=SkillVersion,
        )

    async def delete(
        self,
        version: str,
        *,
        skill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> DeletedSkillVersion:
        """
        Delete a skill version.

        Args:
          version: The skill version number.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not skill_id:
            raise ValueError(f"Expected a non-empty value for `skill_id` but received {skill_id!r}")
        if not version:
            raise ValueError(f"Expected a non-empty value for `version` but received {version!r}")
        return await self._delete(
            path_template("/skills/{skill_id}/versions/{version}", skill_id=skill_id, version=version),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"bearer_auth": True},
            ),
            cast_to=DeletedSkillVersion,
        )


class VersionsWithRawResponse:
    def __init__(self, versions: Versions) -> None:
        self._versions = versions

        self.create = _legacy_response.to_raw_response_wrapper(
            versions.create,
        )
        self.retrieve = _legacy_response.to_raw_response_wrapper(
            versions.retrieve,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            versions.list,
        )
        self.delete = _legacy_response.to_raw_response_wrapper(
            versions.delete,
        )

    @cached_property
    def content(self) -> ContentWithRawResponse:
        return ContentWithRawResponse(self._versions.content)


class AsyncVersionsWithRawResponse:
    def __init__(self, versions: AsyncVersions) -> None:
        self._versions = versions

        self.create = _legacy_response.async_to_raw_response_wrapper(
            versions.create,
        )
        self.retrieve = _legacy_response.async_to_raw_response_wrapper(
            versions.retrieve,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            versions.list,
        )
        self.delete = _legacy_response.async_to_raw_response_wrapper(
            versions.delete,
        )

    @cached_property
    def content(self) -> AsyncContentWithRawResponse:
        return AsyncContentWithRawResponse(self._versions.content)


class VersionsWithStreamingResponse:
    def __init__(self, versions: Versions) -> None:
        self._versions = versions

        self.create = to_streamed_response_wrapper(
            versions.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            versions.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            versions.list,
        )
        self.delete = to_streamed_response_wrapper(
            versions.delete,
        )

    @cached_property
    def content(self) -> ContentWithStreamingResponse:
        return ContentWithStreamingResponse(self._versions.content)


class AsyncVersionsWithStreamingResponse:
    def __init__(self, versions: AsyncVersions) -> None:
        self._versions = versions

        self.create = async_to_streamed_response_wrapper(
            versions.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            versions.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            versions.list,
        )
        self.delete = async_to_streamed_response_wrapper(
            versions.delete,
        )

    @cached_property
    def content(self) -> AsyncContentWithStreamingResponse:
        return AsyncContentWithStreamingResponse(self._versions.content)



from __future__ import annotations

import sys
import json
import threading
from typing import Any, Iterator, cast
from pathlib import Path
from datetime import datetime, timezone, timedelta
from contextlib import contextmanager
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from typing_extensions import override

import httpx
import pytest

from openai import OpenAI
from openai.providers import bedrock

_FUTURE_EXPIRATION = "2099-01-01T00:00:00Z"
_AWS_ENVIRONMENT_NAMES = (
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_SESSION_TOKEN",
    "AWS_SECURITY_TOKEN",
    "AWS_PROFILE",
    "AWS_DEFAULT_PROFILE",
    "AWS_REGION",
    "AWS_DEFAULT_REGION",
    "AWS_CONFIG_FILE",
    "AWS_SHARED_CREDENTIALS_FILE",
    "AWS_CREDENTIAL_FILE",
    "AWS_WEB_IDENTITY_TOKEN_FILE",
    "AWS_ROLE_ARN",
    "AWS_ROLE_SESSION_NAME",
    "AWS_CONTAINER_CREDENTIALS_FULL_URI",
    "AWS_CONTAINER_CREDENTIALS_RELATIVE_URI",
    "AWS_CONTAINER_AUTHORIZATION_TOKEN",
    "AWS_CONTAINER_AUTHORIZATION_TOKEN_FILE",
    "AWS_EC2_METADATA_DISABLED",
    "AWS_EC2_METADATA_SERVICE_ENDPOINT",
    "AWS_EC2_METADATA_SERVICE_ENDPOINT_MODE",
    "AWS_BEARER_TOKEN_BEDROCK",
    "AWS_BEDROCK_BASE_URL",
    "BOTO_CONFIG",
    "HTTP_PROXY",
    "HTTPS_PROXY",
    "ALL_PROXY",
    "http_proxy",
    "https_proxy",
    "all_proxy",
    "NO_PROXY",
    "no_proxy",
)


def _isolate_aws_environment(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> tuple[Path, Path]:
    for name in _AWS_ENVIRONMENT_NAMES:
        monkeypatch.delenv(name, raising=False)

    config_path = tmp_path / "config"
    credentials_path = tmp_path / "credentials"
    config_path.write_text("")
    credentials_path.write_text("")
    monkeypatch.setenv("AWS_CONFIG_FILE", str(config_path))
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", str(credentials_path))
    monkeypatch.setenv("AWS_EC2_METADATA_DISABLED", "true")
    monkeypatch.setenv("NO_PROXY", "127.0.0.1,localhost")
    monkeypatch.setenv("no_proxy", "127.0.0.1,localhost")
    return config_path, credentials_path


def _signed_request(**provider_options: Any) -> httpx.Request:
    requests: list[httpx.Request] = []

    def handler(request: httpx.Request) -> httpx.Response:
        requests.append(request)
        return httpx.Response(200, request=request, json={})

    with OpenAI(
        provider=bedrock(**provider_options),
        http_client=httpx.Client(transport=httpx.MockTransport(handler), trust_env=False),
    ) as client:
        client.get("/models", cast_to=httpx.Response)

    assert len(requests) == 1
    return requests[0]


def _credentials_metadata(name: str, *, expiration: str = _FUTURE_EXPIRATION) -> dict[str, str]:
    return {
        "access_key": f"{name}-access-key",
        "secret_key": f"{name}-secret-key",
        "token": f"{name}-session-token",
        "expiry_time": expiration,
    }


def _assert_signed_with(request: httpx.Request, name: str, *, region: str) -> None:
    assert request.url.host == f"bedrock-mantle.{region}.api.aws"
    assert f"Credential={name}-access-key/" in request.headers["Authorization"]
    assert request.headers["X-Amz-Security-Token"] == f"{name}-session-token"


@contextmanager
def _metadata_server() -> Iterator[tuple[str, list[tuple[str, str, str | None]]]]:
    calls: list[tuple[str, str, str | None]] = []

    class Handler(BaseHTTPRequestHandler):
        def _respond(self, body: str, *, content_type: str = "text/plain") -> None:
            encoded = body.encode()
            self.send_response(200)
            self.send_header("Content-Type", content_type)
            self.send_header("Content-Length", str(len(encoded)))
            self.end_headers()
            self.wfile.write(encoded)

        def do_PUT(self) -> None:
            calls.append(("PUT", self.path, self.headers.get("Authorization")))
            if self.path != "/latest/api/token":
                self.send_error(404)
                return
            self._respond("metadata-token")

        def do_GET(self) -> None:
            calls.append(("GET", self.path, self.headers.get("Authorization")))
            if self.path == "/container-credentials":
                self._respond(
                    json.dumps(
                        {
                            "AccessKeyId": "container-access-key",
                            "SecretAccessKey": "container-secret-key",
                            "Token": "container-session-token",
                            "Expiration": _FUTURE_EXPIRATION,
                        }
                    ),
                    content_type="application/json",
                )
                return
            if self.path == "/latest/meta-data/iam/security-credentials/":
                self._respond("instance-role")
                return
            if self.path == "/latest/meta-data/iam/security-credentials/instance-role":
                self._respond(
                    json.dumps(
                        {
                            "Code": "Success",
                            "LastUpdated": "2026-01-01T00:00:00Z",
                            "Type": "AWS-HMAC",
                            "AccessKeyId": "imds-access-key",
                            "SecretAccessKey": "imds-secret-key",
                            "Token": "imds-session-token",
                            "Expiration": _FUTURE_EXPIRATION,
                        }
                    ),
                    content_type="application/json",
                )
                return
            self.send_error(404)

        @override
        def log_message(self, format: str, *args: Any) -> None:
            del format, args
            return

    server = ThreadingHTTPServer(("127.0.0.1", 0), Handler)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    host, port = cast("tuple[str, int]", server.server_address)
    try:
        yield f"http://{host}:{port}", calls
    finally:
        server.shutdown()
        server.server_close()
        thread.join()


def test_default_chain_uses_environment_session_credentials(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    _isolate_aws_environment(monkeypatch, tmp_path)
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "environment-access-key")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "environment-secret-key")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "environment-session-token")
    monkeypatch.setenv("AWS_REGION", "us-east-1")

    request = _signed_request()

    _assert_signed_with(request, "environment", region="us-east-1")


def test_default_chain_uses_aws_profile_and_shared_files(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path, credentials_path = _isolate_aws_environment(monkeypatch, tmp_path)
    credentials_path.write_text(
        "[work]\n"
        "aws_access_key_id = profile-access-key\n"
        "aws_secret_access_key = profile-secret-key\n"
        "aws_session_token = profile-session-token\n"
    )
    config_path.write_text("[profile work]\nregion = us-west-2\n")
    monkeypatch.setenv("AWS_PROFILE", "work")

    request = _signed_request()

    _assert_signed_with(request, "profile", region="us-west-2")


def test_default_chain_uses_assume_role_profile(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path, credentials_path = _isolate_aws_environment(monkeypatch, tmp_path)
    credentials_path.write_text(
        "[source]\naws_access_key_id = source-access-key\naws_secret_access_key = source-secret-key\n"
    )
    config_path.write_text(
        "[profile assumed]\n"
        "role_arn = arn:aws:iam::123456789012:role/example\n"
        "source_profile = source\n"
        "region = us-east-2\n"
    )
    monkeypatch.setenv("AWS_PROFILE", "assumed")
    credentials_module = pytest.importorskip("botocore.credentials")
    fetches = 0

    def fetch_credentials(_: object) -> dict[str, str]:
        nonlocal fetches
        fetches += 1
        return _credentials_metadata("assume-role")

    monkeypatch.setattr(credentials_module.AssumeRoleCredentialFetcher, "fetch_credentials", fetch_credentials)

    request = _signed_request()

    assert fetches == 1
    _assert_signed_with(request, "assume-role", region="us-east-2")


def test_default_chain_uses_sso_profile(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path, _ = _isolate_aws_environment(monkeypatch, tmp_path)
    config_path.write_text(
        "[profile sso]\n"
        "sso_session = company\n"
        "sso_account_id = 123456789012\n"
        "sso_role_name = Example\n"
        "region = us-west-1\n\n"
        "[sso-session company]\n"
        "sso_start_url = https://example.awsapps.com/start\n"
        "sso_region = us-east-1\n"
    )
    monkeypatch.setenv("AWS_PROFILE", "sso")
    credentials_module = pytest.importorskip("botocore.credentials")
    fetches = 0

    def fetch_credentials(_: object) -> dict[str, str]:
        nonlocal fetches
        fetches += 1
        return _credentials_metadata("sso")

    monkeypatch.setattr(credentials_module.SSOCredentialFetcher, "fetch_credentials", fetch_credentials)

    request = _signed_request()

    assert fetches == 1
    _assert_signed_with(request, "sso", region="us-west-1")


def test_default_chain_uses_web_identity(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    _isolate_aws_environment(monkeypatch, tmp_path)
    token_path = tmp_path / "web-identity-token"
    token_path.write_text("web-identity-token")
    monkeypatch.setenv("AWS_ROLE_ARN", "arn:aws:iam::123456789012:role/web-identity")
    monkeypatch.setenv("AWS_ROLE_SESSION_NAME", "bedrock-test")
    monkeypatch.setenv("AWS_WEB_IDENTITY_TOKEN_FILE", str(token_path))
    monkeypatch.setenv("AWS_REGION", "eu-west-1")
    credentials_module = pytest.importorskip("botocore.credentials")
    fetches = 0

    def fetch_credentials(_: object) -> dict[str, str]:
        nonlocal fetches
        fetches += 1
        return _credentials_metadata("web-identity")

    monkeypatch.setattr(
        credentials_module.AssumeRoleWithWebIdentityCredentialFetcher,
        "fetch_credentials",
        fetch_credentials,
    )

    request = _signed_request()

    assert fetches == 1
    _assert_signed_with(request, "web-identity", region="eu-west-1")


def test_default_chain_uses_credential_process(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path, _ = _isolate_aws_environment(monkeypatch, tmp_path)
    process_path = tmp_path / "credentials_process.py"
    process_output = {
        "Version": 1,
        "AccessKeyId": "process-access-key",
        "SecretAccessKey": "process-secret-key",
        "SessionToken": "process-session-token",
        "Expiration": _FUTURE_EXPIRATION,
    }
    process_path.write_text(f"print({json.dumps(process_output)!r})\n")
    command = f"{json.dumps(sys.executable)} {json.dumps(str(process_path))}"
    config_path.write_text(f"[profile process]\nregion = ap-southeast-2\ncredential_process = {command}\n")
    monkeypatch.setenv("AWS_PROFILE", "process")

    request = _signed_request()

    _assert_signed_with(request, "process", region="ap-southeast-2")


@pytest.mark.parametrize("use_token_file", [False, True], ids=["ecs", "eks-pod-identity"])
def test_default_chain_uses_container_credentials(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, use_token_file: bool
) -> None:
    _isolate_aws_environment(monkeypatch, tmp_path)
    monkeypatch.setenv("AWS_REGION", "us-west-2")
    with _metadata_server() as (base_url, metadata_calls):
        monkeypatch.setenv("AWS_CONTAINER_CREDENTIALS_FULL_URI", f"{base_url}/container-credentials")
        expected_authorization = None
        if use_token_file:
            token_path = tmp_path / "container-authorization-token"
            token_path.write_text("pod-identity-token")
            monkeypatch.setenv("AWS_CONTAINER_AUTHORIZATION_TOKEN_FILE", str(token_path))
            expected_authorization = "pod-identity-token"

        request = _signed_request()

    assert metadata_calls == [("GET", "/container-credentials", expected_authorization)]
    _assert_signed_with(request, "container", region="us-west-2")


def test_default_chain_uses_ec2_instance_metadata(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    _isolate_aws_environment(monkeypatch, tmp_path)
    monkeypatch.setenv("AWS_REGION", "us-east-1")
    monkeypatch.setenv("AWS_EC2_METADATA_DISABLED", "false")
    with _metadata_server() as (base_url, metadata_calls):
        monkeypatch.setenv("AWS_EC2_METADATA_SERVICE_ENDPOINT", base_url)

        request = _signed_request()

    assert [(method, path) for method, path, _ in metadata_calls] == [
        ("PUT", "/latest/api/token"),
        ("GET", "/latest/meta-data/iam/security-credentials/"),
        ("GET", "/latest/meta-data/iam/security-credentials/instance-role"),
    ]
    _assert_signed_with(request, "imds", region="us-east-1")


def test_default_chain_refreshes_credentials_before_retry(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    _isolate_aws_environment(monkeypatch, tmp_path)
    monkeypatch.setenv("AWS_REGION", "us-east-1")
    credentials_module = pytest.importorskip("botocore.credentials")
    session_module = pytest.importorskip("botocore.session")
    refreshes = 0

    def expires_soon() -> str:
        return (datetime.now(timezone.utc) + timedelta(minutes=5)).isoformat()

    def refresh() -> dict[str, str]:
        nonlocal refreshes
        refreshes += 1
        return _credentials_metadata(f"refresh-{refreshes}", expiration=expires_soon())

    credentials = credentials_module.RefreshableCredentials.create_from_metadata(
        metadata=_credentials_metadata("initial", expiration=expires_soon()),
        refresh_using=refresh,
        method="assume-role",
    )

    def get_credentials(_: object) -> Any:
        return credentials

    monkeypatch.setattr(session_module.Session, "get_credentials", get_credentials)
    requests: list[httpx.Request] = []

    def handler(request: httpx.Request) -> httpx.Response:
        requests.append(request)
        return httpx.Response(500 if len(requests) == 1 else 200, request=request, json={})

    with OpenAI(
        provider=bedrock(),
        max_retries=1,
        http_client=httpx.Client(transport=httpx.MockTransport(handler), trust_env=False),
    ) as client:
        client.get("/models", cast_to=httpx.Response)

    assert refreshes == 2
    assert len(requests) == 2
    _assert_signed_with(requests[0], "refresh-1", region="us-east-1")
    _assert_signed_with(requests[1], "refresh-2", region="us-east-1")



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import TypedDict

__all__ = ["RateLimitListRateLimitsParams"]


class RateLimitListRateLimitsParams(TypedDict, total=False):
    after: str
    """A cursor for use in pagination.

    `after` is an object ID that defines your place in the list. For instance, if
    you make a list request and receive 100 objects, ending with obj_foo, your
    subsequent call can include after=obj_foo in order to fetch the next page of the
    list.
    """

    before: str
    """A cursor for use in pagination.

    `before` is an object ID that defines your place in the list. For instance, if
    you make a list request and receive 100 objects, beginning with obj_foo, your
    subsequent call can include before=obj_foo in order to fetch the previous page
    of the list.
    """

    limit: int
    """A limit on the number of objects to be returned. The default is 100."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ApplyPatchTool"]


class ApplyPatchTool(BaseModel):
    """Allows the assistant to create, delete, or update files using unified diffs."""

    type: Literal["apply_patch"]
    """The type of the tool. Always `apply_patch`."""

    allowed_callers: Optional[List[Literal["direct", "programmatic"]]] = None
    """The tool invocation context(s)."""



from pathlib import Path

import anyio
import pytest
from dirty_equals import IsDict, IsList, IsBytes, IsTuple

from openai._files import to_httpx_files, deepcopy_with_paths, async_to_httpx_files
from openai._utils import extract_files

readme_path = Path(__file__).parent.parent.joinpath("README.md")


def test_pathlib_includes_file_name() -> None:
    result = to_httpx_files({"file": readme_path})
    print(result)
    assert result == IsDict({"file": IsTuple("README.md", IsBytes())})


def test_tuple_input() -> None:
    result = to_httpx_files([("file", readme_path)])
    print(result)
    assert result == IsList(IsTuple("file", IsTuple("README.md", IsBytes())))


@pytest.mark.asyncio
async def test_async_pathlib_includes_file_name() -> None:
    result = await async_to_httpx_files({"file": readme_path})
    print(result)
    assert result == IsDict({"file": IsTuple("README.md", IsBytes())})


@pytest.mark.asyncio
async def test_async_supports_anyio_path() -> None:
    result = await async_to_httpx_files({"file": anyio.Path(readme_path)})
    print(result)
    assert result == IsDict({"file": IsTuple("README.md", IsBytes())})


@pytest.mark.asyncio
async def test_async_tuple_input() -> None:
    result = await async_to_httpx_files([("file", readme_path)])
    print(result)
    assert result == IsList(IsTuple("file", IsTuple("README.md", IsBytes())))


def test_string_not_allowed() -> None:
    with pytest.raises(TypeError, match="Expected file types input to be a FileContent type or to be a tuple"):
        to_httpx_files(
            {
                "file": "foo",  # type: ignore
            }
        )


def assert_different_identities(obj1: object, obj2: object) -> None:
    assert obj1 == obj2
    assert obj1 is not obj2


class TestDeepcopyWithPaths:
    def test_copies_top_level_dict(self) -> None:
        original = {"file": b"data", "other": "value"}
        result = deepcopy_with_paths(original, [["file"]])
        assert_different_identities(result, original)

    def test_file_value_is_same_reference(self) -> None:
        file_bytes = b"contents"
        original = {"file": file_bytes}
        result = deepcopy_with_paths(original, [["file"]])
        assert_different_identities(result, original)
        assert result["file"] is file_bytes

    def test_list_popped_wholesale(self) -> None:
        files = [b"f1", b"f2"]
        original = {"files": files, "title": "t"}
        result = deepcopy_with_paths(original, [["files", "<array>"]])
        assert_different_identities(result, original)
        result_files = result["files"]
        assert isinstance(result_files, list)
        assert_different_identities(result_files, files)

    def test_nested_array_path_copies_list_and_elements(self) -> None:
        elem1 = {"file": b"f1", "extra": 1}
        elem2 = {"file": b"f2", "extra": 2}
        original = {"items": [elem1, elem2]}
        result = deepcopy_with_paths(original, [["items", "<array>", "file"]])
        assert_different_identities(result, original)
        result_items = result["items"]
        assert isinstance(result_items, list)
        assert_different_identities(result_items, original["items"])
        assert_different_identities(result_items[0], elem1)
        assert_different_identities(result_items[1], elem2)

    def test_empty_paths_returns_same_object(self) -> None:
        original = {"foo": "bar"}
        result = deepcopy_with_paths(original, [])
        assert result is original

    def test_multiple_paths(self) -> None:
        f1 = b"file1"
        f2 = b"file2"
        original = {"a": f1, "b": f2, "c": "unchanged"}
        result = deepcopy_with_paths(original, [["a"], ["b"]])
        assert_different_identities(result, original)
        assert result["a"] is f1
        assert result["b"] is f2
        assert result["c"] is original["c"]

    def test_extract_files_does_not_mutate_original_top_level(self) -> None:
        file_bytes = b"contents"
        original = {"file": file_bytes, "other": "value"}

        copied = deepcopy_with_paths(original, [["file"]])
        extracted = extract_files(copied, paths=[["file"]])

        assert extracted == [("file", file_bytes)]
        assert original == {"file": file_bytes, "other": "value"}
        assert copied == {"other": "value"}

    def test_extract_files_does_not_mutate_original_nested_array_path(self) -> None:
        file1 = b"f1"
        file2 = b"f2"
        original = {
            "items": [
                {"file": file1, "extra": 1},
                {"file": file2, "extra": 2},
            ],
            "title": "example",
        }

        copied = deepcopy_with_paths(original, [["items", "<array>", "file"]])
        extracted = extract_files(copied, paths=[["items", "<array>", "file"]])

        assert [entry for _, entry in extracted] == [file1, file2]
        assert original == {
            "items": [
                {"file": file1, "extra": 1},
                {"file": file2, "extra": 2},
            ],
            "title": "example",
        }
        assert copied == {
            "items": [
                {"extra": 1},
                {"extra": 2},
            ],
            "title": "example",
        }



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["AdminAPIKeyCreateParams"]


class AdminAPIKeyCreateParams(TypedDict, total=False):
    name: Required[str]

    expires_in_seconds: int
    """The number of seconds until the API key expires.

    Omit this field for a key that does not expire.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..responses.response_output_refusal import ResponseOutputRefusal

__all__ = ["RefusalContent"]

RefusalContent = ResponseOutputRefusal



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..responses.response_output_text import ResponseOutputText

__all__ = ["OutputTextContent"]

OutputTextContent = ResponseOutputText



from __future__ import annotations

import sys
import asyncio
import warnings
import threading
import subprocess
import importlib.util
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from typing_extensions import override

import httpx
import respx
import pytest

import openai
from openai import OpenAI, AsyncOpenAI, _httpx2 as httpx2_helpers


def test_base_import_does_not_load_httpx2() -> None:
    subprocess.run([sys.executable, "-c", "import sys; import openai; assert 'httpx2' not in sys.modules"], check=True)


def test_missing_httpx2_extra_is_actionable(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(httpx2_helpers.sys, "version_info", (3, 10))

    def missing_httpx2(name: str) -> object:
        raise ImportError(name)

    monkeypatch.setattr(httpx2_helpers.importlib, "import_module", missing_httpx2)

    for helper in (openai.DefaultHttpx2Client, openai.DefaultAsyncHttpx2Client):
        with pytest.raises(RuntimeError, match=r"install the httpx2 extra: pip install 'openai\[httpx2\]'"):
            helper()


def test_python39_httpx2_error_is_actionable(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(httpx2_helpers.sys, "version_info", (3, 9))

    for helper in (openai.DefaultHttpx2Client, openai.DefaultAsyncHttpx2Client):
        with pytest.raises(RuntimeError, match=r"HTTPX2 requires Python 3\.10 or later.*openai\[httpx2\]"):
            helper()


@pytest.mark.respx(base_url="https://example.test/v1")
def test_default_httpx_family_and_respx_are_unchanged(respx_mock: respx.MockRouter) -> None:
    route = respx_mock.get("/models").mock(return_value=httpx.Response(200, json={"object": "list", "data": []}))

    with warnings.catch_warnings(record=True) as captured:
        warnings.simplefilter("always")
        with OpenAI(api_key="test", base_url="https://example.test/v1") as client:
            response = client.models.with_raw_response.list()

    assert route.called
    assert captured == []
    assert isinstance(response.http_response, httpx.Response)
    assert isinstance(response.http_request, httpx.Request)


async def test_existing_httpx_helpers_and_injected_clients_are_unchanged() -> None:
    class SyncHttpxClient(openai.DefaultHttpxClient):
        pass

    class AsyncHttpxClient(openai.DefaultAsyncHttpxClient):
        pass

    def handler(request: httpx.Request) -> httpx.Response:
        return httpx.Response(200, json={"object": "list", "data": []}, request=request)

    with SyncHttpxClient(transport=httpx.MockTransport(handler), trust_env=False) as http_client:
        with OpenAI(api_key="test", base_url="https://example.test/v1", http_client=http_client) as client:
            response = client.models.with_raw_response.list()
            assert isinstance(response.http_response, httpx.Response)

    async with AsyncHttpxClient(transport=httpx.MockTransport(handler), trust_env=False) as http_client:
        async with AsyncOpenAI(api_key="test", base_url="https://example.test/v1", http_client=http_client) as client:
            response = await client.models.with_raw_response.list()
            assert isinstance(response.http_response, httpx.Response)


async def test_existing_aiohttp_adapter_is_unchanged_when_installed() -> None:
    if importlib.util.find_spec("httpx_aiohttp") is None:
        pytest.skip("the aiohttp extra is not installed")

    class ModelsHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            self.send_response(200)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"object":"list","data":[]}')

        @override
        def log_message(self, format: str, *_args: object) -> None:  # noqa: A002
            return None

    server = ThreadingHTTPServer(("127.0.0.1", 0), ModelsHandler)
    thread = threading.Thread(target=server.serve_forever)
    thread.start()
    try:
        async with AsyncOpenAI(
            api_key="test",
            base_url=f"http://127.0.0.1:{server.server_port}/v1",
            http_client=openai.DefaultAioHttpClient(),
            max_retries=0,
        ) as client:
            response = await client.models.with_raw_response.list()
    finally:
        await asyncio.to_thread(server.shutdown)
        thread.join()
        server.server_close()

    assert isinstance(response.http_response, httpx.Response)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import Literal, Required, TypedDict

from .tool_param import ToolParam

__all__ = ["ResponseToolSearchOutputItemParamParam"]


class ResponseToolSearchOutputItemParamParam(TypedDict, total=False):
    tools: Required[Iterable[ToolParam]]
    """The loaded tool definitions returned by the tool search output."""

    type: Required[Literal["tool_search_output"]]
    """The item type. Always `tool_search_output`."""

    id: Optional[str]
    """The unique ID of this tool search output."""

    call_id: Optional[str]
    """The unique ID of the tool search call generated by the model."""

    execution: Literal["server", "client"]
    """Whether tool search was executed by the server or by the client."""

    status: Optional[Literal["in_progress", "completed", "incomplete"]]
    """The status of the tool search output."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Union, Optional
from typing_extensions import Literal

from ..._models import BaseModel
from ..file_chunking_strategy import FileChunkingStrategy

__all__ = ["VectorStoreFile", "LastError"]


class LastError(BaseModel):
    """The last error associated with this vector store file.

    Will be `null` if there are no errors.
    """

    code: Literal["server_error", "unsupported_file", "invalid_file"]
    """One of `server_error`, `unsupported_file`, or `invalid_file`."""

    message: str
    """A human-readable description of the error."""


class VectorStoreFile(BaseModel):
    """A list of files attached to a vector store."""

    id: str
    """The identifier, which can be referenced in API endpoints."""

    created_at: int
    """The Unix timestamp (in seconds) for when the vector store file was created."""

    last_error: Optional[LastError] = None
    """The last error associated with this vector store file.

    Will be `null` if there are no errors.
    """

    object: Literal["vector_store.file"]
    """The object type, which is always `vector_store.file`."""

    status: Literal["in_progress", "completed", "cancelled", "failed"]
    """
    The status of the vector store file, which can be either `in_progress`,
    `completed`, `cancelled`, or `failed`. The status `completed` indicates that the
    vector store file is ready for use.
    """

    usage_bytes: int
    """The total vector store usage in bytes.

    Note that this may be different from the original file size.
    """

    vector_store_id: str
    """
    The ID of the
    [vector store](https://platform.openai.com/docs/api-reference/vector-stores/object)
    that the [File](https://platform.openai.com/docs/api-reference/files) is
    attached to.
    """

    attributes: Optional[Dict[str, Union[str, float, bool]]] = None
    """Set of 16 key-value pairs that can be attached to an object.

    This can be useful for storing additional information about the object in a
    structured format, and querying for objects via API or the dashboard. Keys are
    strings with a maximum length of 64 characters. Values are strings with a
    maximum length of 512 characters, booleans, or numbers.
    """

    chunking_strategy: Optional[FileChunkingStrategy] = None
    """The strategy used to chunk the file."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["BetaResponseInjectCreatedEvent"]


class BetaResponseInjectCreatedEvent(BaseModel):
    """
    Emitted when all injected input items were validated and committed to the
    active response.
    """

    response_id: str
    """The ID of the response that accepted the input."""

    sequence_number: int
    """The sequence number for this event."""

    type: Literal["response.inject.created"]
    """The event discriminator. Always `response.inject.created`."""

    stream_id: Optional[str] = None
    """The multiplexed WebSocket stream that emitted the event.

    This field is present only when WebSocket multiplexing is enabled separately.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional

import httpx

from ...... import _legacy_response
from .roles import (
    Roles,
    AsyncRoles,
    RolesWithRawResponse,
    AsyncRolesWithRawResponse,
    RolesWithStreamingResponse,
    AsyncRolesWithStreamingResponse,
)
from ......_types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ......_utils import path_template, maybe_transform, async_maybe_transform
from ......_compat import cached_property
from ......_resource import SyncAPIResource, AsyncAPIResource
from ......_response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ......pagination import SyncConversationCursorPage, AsyncConversationCursorPage
from ......_base_client import AsyncPaginator, make_request_options
from ......types.admin.organization.projects import user_list_params, user_create_params, user_update_params
from ......types.admin.organization.projects.project_user import ProjectUser
from ......types.admin.organization.projects.user_delete_response import UserDeleteResponse

__all__ = ["Users", "AsyncUsers"]


class Users(SyncAPIResource):
    @cached_property
    def roles(self) -> Roles:
        return Roles(self._client)

    @cached_property
    def with_raw_response(self) -> UsersWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return UsersWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> UsersWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return UsersWithStreamingResponse(self)

    def create(
        self,
        project_id: str,
        *,
        role: str,
        email: Optional[str] | Omit = omit,
        user_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """Adds a user to the project.

        Users must already be members of the organization to
        be added to a project.

        Args:
          role: `owner` or `member`

          email: Email of the user to add.

          user_id: The ID of the user.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        return self._post(
            path_template("/organization/projects/{project_id}/users", project_id=project_id),
            body=maybe_transform(
                {
                    "role": role,
                    "email": email,
                    "user_id": user_id,
                },
                user_create_params.UserCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    def retrieve(
        self,
        user_id: str,
        *,
        project_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """
        Retrieves a user in the project.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return self._get(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    def update(
        self,
        user_id: str,
        *,
        project_id: str,
        role: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """
        Modifies a user's role in the project.

        Args:
          role: `owner` or `member`

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return self._post(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            body=maybe_transform({"role": role}, user_update_params.UserUpdateParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    def list(
        self,
        project_id: str,
        *,
        after: str | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncConversationCursorPage[ProjectUser]:
        """
        Returns a list of users in the project.

        Args:
          after: A cursor for use in pagination. `after` is an object ID that defines your place
              in the list. For instance, if you make a list request and receive 100 objects,
              ending with obj_foo, your subsequent call can include after=obj_foo in order to
              fetch the next page of the list.

          limit: A limit on the number of objects to be returned. Limit can range between 1 and
              100, and the default is 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        return self._get_api_list(
            path_template("/organization/projects/{project_id}/users", project_id=project_id),
            page=SyncConversationCursorPage[ProjectUser],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    user_list_params.UserListParams,
                ),
                security={"admin_api_key_auth": True},
            ),
            model=ProjectUser,
        )

    def delete(
        self,
        user_id: str,
        *,
        project_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> UserDeleteResponse:
        """
        Deletes a user from the project.

        Returns confirmation of project user deletion, or an error if the project is
        archived (archived projects have no users).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return self._delete(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=UserDeleteResponse,
        )


class AsyncUsers(AsyncAPIResource):
    @cached_property
    def roles(self) -> AsyncRoles:
        return AsyncRoles(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncUsersWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/openai/openai-python#accessing-raw-response-data-eg-headers
        """
        return AsyncUsersWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncUsersWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/openai/openai-python#with_streaming_response
        """
        return AsyncUsersWithStreamingResponse(self)

    async def create(
        self,
        project_id: str,
        *,
        role: str,
        email: Optional[str] | Omit = omit,
        user_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """Adds a user to the project.

        Users must already be members of the organization to
        be added to a project.

        Args:
          role: `owner` or `member`

          email: Email of the user to add.

          user_id: The ID of the user.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        return await self._post(
            path_template("/organization/projects/{project_id}/users", project_id=project_id),
            body=await async_maybe_transform(
                {
                    "role": role,
                    "email": email,
                    "user_id": user_id,
                },
                user_create_params.UserCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    async def retrieve(
        self,
        user_id: str,
        *,
        project_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """
        Retrieves a user in the project.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return await self._get(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    async def update(
        self,
        user_id: str,
        *,
        project_id: str,
        role: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProjectUser:
        """
        Modifies a user's role in the project.

        Args:
          role: `owner` or `member`

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return await self._post(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            body=await async_maybe_transform({"role": role}, user_update_params.UserUpdateParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=ProjectUser,
        )

    def list(
        self,
        project_id: str,
        *,
        after: str | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[ProjectUser, AsyncConversationCursorPage[ProjectUser]]:
        """
        Returns a list of users in the project.

        Args:
          after: A cursor for use in pagination. `after` is an object ID that defines your place
              in the list. For instance, if you make a list request and receive 100 objects,
              ending with obj_foo, your subsequent call can include after=obj_foo in order to
              fetch the next page of the list.

          limit: A limit on the number of objects to be returned. Limit can range between 1 and
              100, and the default is 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        return self._get_api_list(
            path_template("/organization/projects/{project_id}/users", project_id=project_id),
            page=AsyncConversationCursorPage[ProjectUser],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "after": after,
                        "limit": limit,
                    },
                    user_list_params.UserListParams,
                ),
                security={"admin_api_key_auth": True},
            ),
            model=ProjectUser,
        )

    async def delete(
        self,
        user_id: str,
        *,
        project_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> UserDeleteResponse:
        """
        Deletes a user from the project.

        Returns confirmation of project user deletion, or an error if the project is
        archived (archived projects have no users).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not project_id:
            raise ValueError(f"Expected a non-empty value for `project_id` but received {project_id!r}")
        if not user_id:
            raise ValueError(f"Expected a non-empty value for `user_id` but received {user_id!r}")
        return await self._delete(
            path_template(
                "/organization/projects/{project_id}/users/{user_id}", project_id=project_id, user_id=user_id
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                security={"admin_api_key_auth": True},
            ),
            cast_to=UserDeleteResponse,
        )


class UsersWithRawResponse:
    def __init__(self, users: Users) -> None:
        self._users = users

        self.create = _legacy_response.to_raw_response_wrapper(
            users.create,
        )
        self.retrieve = _legacy_response.to_raw_response_wrapper(
            users.retrieve,
        )
        self.update = _legacy_response.to_raw_response_wrapper(
            users.update,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            users.list,
        )
        self.delete = _legacy_response.to_raw_response_wrapper(
            users.delete,
        )

    @cached_property
    def roles(self) -> RolesWithRawResponse:
        return RolesWithRawResponse(self._users.roles)


class AsyncUsersWithRawResponse:
    def __init__(self, users: AsyncUsers) -> None:
        self._users = users

        self.create = _legacy_response.async_to_raw_response_wrapper(
            users.create,
        )
        self.retrieve = _legacy_response.async_to_raw_response_wrapper(
            users.retrieve,
        )
        self.update = _legacy_response.async_to_raw_response_wrapper(
            users.update,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            users.list,
        )
        self.delete = _legacy_response.async_to_raw_response_wrapper(
            users.delete,
        )

    @cached_property
    def roles(self) -> AsyncRolesWithRawResponse:
        return AsyncRolesWithRawResponse(self._users.roles)


class UsersWithStreamingResponse:
    def __init__(self, users: Users) -> None:
        self._users = users

        self.create = to_streamed_response_wrapper(
            users.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            users.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            users.update,
        )
        self.list = to_streamed_response_wrapper(
            users.list,
        )
        self.delete = to_streamed_response_wrapper(
            users.delete,
        )

    @cached_property
    def roles(self) -> RolesWithStreamingResponse:
        return RolesWithStreamingResponse(self._users.roles)


class AsyncUsersWithStreamingResponse:
    def __init__(self, users: AsyncUsers) -> None:
        self._users = users

        self.create = async_to_streamed_response_wrapper(
            users.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            users.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            users.update,
        )
        self.list = async_to_streamed_response_wrapper(
            users.list,
        )
        self.delete = async_to_streamed_response_wrapper(
            users.delete,
        )

    @cached_property
    def roles(self) -> AsyncRolesWithStreamingResponse:
        return AsyncRolesWithStreamingResponse(self._users.roles)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["BetaResponseContainerReference"]


class BetaResponseContainerReference(BaseModel):
    """Represents a container created with /v1/containers."""

    container_id: str

    type: Literal["container_reference"]
    """The environment type. Always `container_reference`."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional

from ....._models import BaseModel

__all__ = ["RoleListResponse", "AssignmentSource"]


class AssignmentSource(BaseModel):
    principal_id: str

    principal_type: str


class RoleListResponse(BaseModel):
    """
    Detailed information about a role assignment entry returned when listing assignments.
    """

    id: str
    """Identifier for the role."""

    assignment_sources: Optional[List[AssignmentSource]] = None
    """Principals from which the role assignment is inherited, when available."""

    created_at: Optional[int] = None
    """When the role was created."""

    created_by: Optional[str] = None
    """Identifier of the actor who created the role."""

    created_by_user_obj: Optional[Dict[str, object]] = None
    """User details for the actor that created the role, when available."""

    description: Optional[str] = None
    """Description of the role."""

    metadata: Optional[Dict[str, object]] = None
    """Arbitrary metadata stored on the role."""

    name: str
    """Name of the role."""

    permissions: List[str]
    """Permissions associated with the role."""

    predefined_role: bool
    """Whether the role is predefined by OpenAI."""

    resource_type: str
    """Resource type the role applies to."""

    updated_at: Optional[int] = None
    """When the role was last updated."""
